/*

 overheatd 0.1
 by Hanno Boeck, http://www.hboeck.de/

 latest version is available at http://www.int21.de/overheatd/

I, the copyright holder of this work, hereby release it into the public
domain. This applies worldwide.
In case this is not legally possible:
I grant anyone the right to use this work for any purpose, without any
conditions, unless such conditions are required by law.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define MAX_CPUS 16

FILE *logfd;

int get_temperature(int cpu) {
	char sysfsfilename[38];
	char tempstr[10];
	int fd, ret;
	sprintf(sysfsfilename, "/sys/class/thermal/thermal_zone%i/temp", cpu);
	fd = open(sysfsfilename, O_RDONLY);
	ret = read(fd, tempstr, 10);
	if (!ret) {
		fprintf(logfd, "Unable to get temperature, exiting\n");
		exit(EXIT_FAILURE);
	}
	close(fd);
	return atoi(tempstr)/1000;
}

/* cpu: cpu number in sysfs                      */
/* cool: 0 for ondemand, 1 for cooling/powersafe */
void cool_cpu(int cpu, int cool) {
	int fd, ret;
	char filename[54];
	sprintf(filename, "/sys/devices/system/cpu/cpu%i/cpufreq/scaling_governor", cpu);
	fd = open(filename, O_WRONLY);
	if (cool == 1) ret = write(fd, "powersave", 9);
	else ret = write(fd, "ondemand", 8);
	if (!ret) {
		fprintf(logfd, "Unable to set governor, exiting\n");
		exit(EXIT_FAILURE);
	}
	close(fd);
}

int main(int argc, char **argv) {
	int i, fd, cpul, cpuh, temp, ret;
	char s[5];
	char filename[54];
	char tempstr[9];
	int hot[MAX_CPUS];
	pid_t pid, sid;
	char c;
	unsigned int critical = 90;
	char *logfilename = "/var/log/overheatd.log";
	unsigned int poll = 5;

	while ((c = getopt(argc, argv, "t:p:l:")) != -1)
		switch (c) {
		case 't':
			critical = atoi(optarg);
			break;
		case 'p':
			poll = atoi(optarg);
			break;
		case 'l':
			logfilename = optarg;
			break;
		case '?':
			printf("overheatd - daemon to slowdown cpu on overheating\n");
			printf("\n");
			printf("-t [temp]\tSet temperature treshold in °C (default: 90)\n");
			printf("-p [poll]\tpoll interval in seconds (default: 5)\n");
			printf("-l [logfile]\tlogfile (default: /var/log/overheatd.log)\n");
			printf("\n");
			exit(EXIT_SUCCESS);
		}

	umask(0077);

	logfd = fopen(logfilename, "a");

	/* Get available CPUs */
	fd = open("/sys/devices/system/cpu/online", O_RDONLY);
	ret = read(fd, s, 5);
	if (!ret) {
		fprintf(logfd, "Failed to read /sys/devices/system/cpu/online\n");
		exit(EXIT_FAILURE);
	}
	cpul = atoi(strtok(s, "-"));
	cpuh = atoi(strtok(NULL, "-"));
	close(fd);
	fprintf(logfd, "Initializing overheatd: cpu %i to %i\n", cpul, cpuh);

	/* initializing, checking for  */
	for (i = cpul; i <= cpuh; i++) {
		hot[i]=0;
		sprintf(filename, "/sys/devices/system/cpu/cpu%i/cpufreq/scaling_governor", i);
		fd = open(filename, O_RDONLY);
		ret = read(fd, tempstr, 8);
		if (!ret) {
			fprintf(logfd, "Failed to read %s\n", filename);
			exit(EXIT_FAILURE);
		}
		close(fd);
		if (strncmp(tempstr, "ondemand", 8)) {
			fprintf(logfd, "setting cpu %i to ondemand\n", i);
			cool_cpu(i, 0);
		}
	}

	/* Fork off the parent process */
	pid = fork();
	if (pid < 0) exit(EXIT_FAILURE);
	if (pid > 0) exit(EXIT_SUCCESS);
	sid = setsid();
	if (sid < 0) exit(EXIT_FAILURE);

	if ((chdir("/")) < 0) exit(EXIT_FAILURE);

	/* Close standard file descriptors */
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	while(1) {
		fflush(logfd);
		sleep(poll);

		for (i = cpul; i <= cpuh; i++) {
			temp = get_temperature(i);
			if ((hot[i]==0) && (temp > critical)) {
				fprintf(logfd, "ALERT! Temperature on CPU %i is very high, %i °C\n", i, temp);
				cool_cpu(i, 1);
				hot[i] = 1;
			} else if ((hot[i]==1) && (temp <= critical)) {
				fprintf(logfd, "CPU %i back to normal, %i °C\n", i, temp);
				cool_cpu(i, 0);
				hot[i] = 0;
			}
		}
	}
}
